"""
This script takes a DuckDB database with a buildings table and converts it to GeoParquet 
files partitioned on first country and then quadkey. The buildings table must have a
country_iso field and quadkey field, populated by overture-buildings-parquet-add-columns.py.
The main function is process_db(), and it will take as input a maximum number of rows per
file and a row group size for the Parquet files. It will then iterate through the countries
in the database and partition the buildings table into GeoParquet files for each country.
If the number of rows for a country is greater than the maximum number of rows per file,
it will partition the country into quadkeys and create GeoParquet files for each quadkey.
Those quadkeys will be further partitioned if necessary until the number of rows for a
quadkey is less than or equal to the maximum number of rows per file. 
"""

import duckdb
import datetime
import subprocess
import tempfile
import os
import click
import shutil
import geopandas as gpd
from shapely import wkb
import pandas as pd
import time

def current_time_str():
    return datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')

def print_verbose(msg, verbose):
    if verbose:
        print(f"[{current_time_str()}] {msg}")

def convert_gpq(input_filename, row_group_size, verbose):
    print_verbose(f"Starting conversion for {input_filename} using gpq (row_group_size ignored).", verbose)

    # Create a temporary file
    temp_file = tempfile.NamedTemporaryFile(suffix=".parquet", delete=False)
    temp_file.close()  # Close the file so gpq can open it

    # Convert the Parquet file to a GeoParquet file using gpq
    gpq_cmd = ['gpq', 'convert', input_filename, temp_file.name]
    subprocess.run(gpq_cmd, check=True)

    print_verbose(f"Conversion for {input_filename} using gpq finished.", verbose)

    # Rename (move) the temp file to the final filename
    shutil.move(temp_file.name, input_filename)

    # Delete the initial temp file if it still exists
    #initial_temp_filename = f'{country_code}_temp.parquet'
    #if os.path.exists(initial_temp_filename):
    #    os.remove(initial_temp_filename)

def convert_pandas(input_filename, rg_size, verbose):
    # Placeholder function to be fleshed out
    print_verbose("Starting conversion using pandas.", verbose)
    try:
        df = pd.read_parquet(input_filename)

        # Convert WKB geometry to geopandas geometry
        df['geometry'] = df['geometry'].apply(wkb.loads, hex=True)
        gdf = gpd.GeoDataFrame(df, geometry='geometry', crs="EPSG:4326")
        # Change output file the input_filename with .parquet replaced with _geo.parquet
        output_filename = input_filename.replace(".parquet", "_geo.parquet")
    
        gdf.to_parquet(output_filename, row_group_size=rg_size)
        # delete the original file
        os.remove(input_filename)
        # Rename (move) the output file to the input filename
        shutil.move(output_filename, input_filename)
        print(f"Finished processing {input_filename} at {time.ctime()}")
    except Exception as e:
        print(f"Error processing {input_filename}: {e}")

#not quite working yet - not sure what's wrong. Should go faster than pandas.  
def convert_ogr(input_filename, rg_size, verbose):
    fields_to_keep = ['confidence', 'area_in_meters', 'full_plus_code', 'country_iso', 'quadkey']
    output_filename = input_filename.replace(".parquet", "_geo.parquet")
    rg_cmd = f"ROW_GROUP_SIZE={rg_size}"
    cmd = [
        'ogr2ogr',
        '-f',
        'Parquet',
        '-select',
        ','.join(fields_to_keep),
        output_filename,
        input_filename,
   #     '-oo',
   #     rg_cmd,
        '-oo',
        'GEOM_POSSIBLE_NAMES=geometry',
        '-a_srs',
        'EPSG:4326', ]

    # print the ogr2ogr command that will be run
    if verbose:
        print("ogr2ogr command:")
        print(' '.join(cmd))

    # Run the command
    subprocess.run(cmd, check=True)

    # delete the original file
    os.remove(input_filename)
    # Rename (move) the output file to the input filename
    shutil.move(output_filename, input_filename)
    print(f"Finished processing {input_filename} at {time.ctime()}")

    if verbose:
        print(f"Converted {input_filename} to {output_filename} using ogr2ogr.")

 

def fetch_quadkeys(conn, table_name, country_code, length, verbose, prev_qk=""):
    query = f"SELECT DISTINCT SUBSTR(quadkey, 1, {length}) FROM {table_name} WHERE country_iso = '{country_code}'"
    if prev_qk:
        query += f" AND SUBSTR(quadkey, 1, {len(prev_qk)}) = '{prev_qk}'"
    print_verbose(f'Executing: {query}', verbose)
    return conn.execute(query).fetchall()

def convert_to_geoparquet(parquet_path, geo_conversion, row_group_size, verbose):
    if geo_conversion == 'gpq':
        convert_gpq(parquet_path, row_group_size, verbose)
        print_verbose(f"File: {parquet_path} written with gpq", verbose)
    elif geo_conversion == 'pandas':
        convert_pandas(parquet_path, row_group_size, verbose)
        print_verbose(f"File: {parquet_path} written with pandas", verbose)
    elif geo_conversion == 'ogr':
        convert_ogr(parquet_path, row_group_size, verbose)
        print_verbose(f"File: {parquet_path} written with ogr", verbose)
    else:
        print_verbose(f"File: {parquet_path} written without converting to GeoParquet", verbose)

#TODO: go all the way into the quad to find the smallest quadkey that contains less than max_per_file rows
def process_quadkey_recursive(conn, table_name, country_code, output_folder, length, geo_conversion, row_group_size, verbose, max_per_file, current_qk=""):
    distinct_quadkeys = fetch_quadkeys(conn, table_name, country_code, length, verbose, current_qk)
    print_verbose(f"The list of quadkeys for country {country_code} and length {length} is {distinct_quadkeys}", verbose)
    #num_distinct_qk = len(distinct_quadkeys)
    for qk in distinct_quadkeys:
        qk_str = qk[0]
        qk_count_query = f"SELECT COUNT(*) FROM {table_name} WHERE country_iso = '{country_code}' AND SUBSTR(quadkey, 1, {length}) = '{qk_str}'"
        print_verbose(f'Executing: {qk_count_query}', verbose)
        qk_count = conn.execute(qk_count_query).fetchone()[0]
        print_verbose(f"Quadkey {qk_str} has {qk_count} rows", verbose)
        if qk_count > max_per_file:
            process_quadkey_recursive(conn, table_name, country_code, output_folder, length + 1, geo_conversion, row_group_size, verbose, max_per_file, qk_str)
        else:
            quad_output_filename = os.path.join(output_folder, f'{country_code}_{qk_str}.parquet')
            if os.path.exists(quad_output_filename):
                print_verbose(f"Output file {quad_output_filename} already exists, skipping...", verbose)
            else:
                copy_cmd = f"COPY (SELECT * FROM {table_name} WHERE country_iso = '{country_code}' AND SUBSTR(quadkey, 1, {length}) = '{qk_str}' ORDER BY quadkey) TO '{quad_output_filename}' WITH (FORMAT PARQUET);"
                print_verbose(f'Executing: {copy_cmd}', verbose)
                conn.execute(copy_cmd)
                convert_to_geoparquet(quad_output_filename, geo_conversion, row_group_size, verbose)


# TODO: add option for 'hive' output (put things in folder)
# TODO: add option to read duckdb path from an environment variable
# TODO: add row group size option (first works with duckdb)

@click.command()
@click.argument('duckdb-path', type=click.Path(exists=True))
@click.option('--output-folder', default=os.getcwd(), type=click.Path(), help='Folder to store the output files')
@click.option('--geo-conversion', default='gpq', type=click.Choice(['gpq', 'none', 'pandas', 'ogr'], case_sensitive=False))
@click.option('--verbose', is_flag=True, default=False, help='Print verbose output')
@click.option('--max-per-file', default=10000000, type=int, help='Maximum number of rows per file')
@click.option('--row-group-size', default=10000, type=int, help='Row group size for Parquet files')
@click.option('--hive', is_flag=True, default=False, help='Output files in Hive format (folder structure)')
def process_db(duckdb_path, output_folder, geo_conversion, verbose, max_per_file, row_group_size, hive):
    table_name = 'buildings'
    # create output folder if it does not exist
    os.makedirs(output_folder, exist_ok=True)
    conn = duckdb.connect(duckdb_path)
    conn.execute('LOAD spatial;')
    cursor = conn.execute('SELECT DISTINCT country_iso FROM buildings')
    countries = cursor.fetchall()
    
    print_verbose(f'Found {len(countries)} unique countries', verbose)
    countries.reverse()
    for country in countries:
        country_code = country[0]
        write_folder = output_folder
        if (hive):
            write_folder = os.path.join(output_folder, f'country_iso={country_code}')
            os.makedirs(write_folder, exist_ok=True)
        output_filename = os.path.join(write_folder, f'{country_code}.parquet')
        if os.path.exists(output_filename):
            print_verbose(f"Output file for country {country_code} already exists, skipping...", verbose)
            continue

        count_query = f"SELECT COUNT(*) FROM {table_name} WHERE country_iso = '{country_code}'"
        print_verbose(f'Executing: {count_query}', verbose)
        count = conn.execute(count_query).fetchone()[0]
        print_verbose(f"Country {country_code} has {count} rows", verbose)

        if count <= max_per_file:
            copy_cmd = f"COPY (SELECT * FROM {table_name} WHERE country_iso = '{country_code}' ORDER BY quadkey) TO '{output_filename}' WITH (FORMAT PARQUET);"
            print_verbose(f'Executing: {copy_cmd}', verbose)
            conn.execute(copy_cmd)
            convert_to_geoparquet(output_filename, geo_conversion, row_group_size, verbose)
        else:
            process_quadkey_recursive(conn, table_name, country_code, write_folder, 1, geo_conversion, row_group_size, verbose, max_per_file)

if __name__ == "__main__":
    process_db()


import os
import subprocess
import time
from datetime import datetime, timedelta
import json

import click
import glob
import duckdb
import pandas as pd
import geopandas as gpd
from shapely import wkt
from shapely.geometry import mapping
from openlocationcode import openlocationcode as olc

# Global variable, that runs GPQ (https://github.com/planetlabs/gpq) after DuckDB writes the Parquet file.
# This is necessary because DuckDB does not write the GeoParquet metadata (yet). Once DuckDB implements
# this feature can be removed. Setting it to false will give a sense of how fast DuckDB will be, but
# if you want to actually use the output GeoParquet files, set it to True.
RUN_GPQ_CONVERSION = True

# Global variable, that sets the compression type for the Parquet files. The two options that
# will work for both DuckDB and pandas are 'snappy' and 'gzip'. 'snappy' is the default. You can
# try out brotli with pandas, it seems to give the most compression. DuckDB additional supports
# zstd, but pandas does not. Note that GPQ conversion on DuckDB output likely keeps the same
# compression, but I have not tested this. GPQ conversion from Parquet does not yet support
# the other GPQ compression options.
PARQUET_COMPRESSION = 'snappy'

# Don't run the DuckDB GPKG conversion if set to true, as it takes a long time, likely due to a bug.
# It means longer runs and puts one big time on the graphs.
SKIP_DUCK_GPKG = True

@click.group()
def cli():
    pass


def define_output_paths(input_file_path, output_directory, format):
    output_file_name = os.path.basename(input_file_path)[:-3] + format
    output_file_path = os.path.join(output_directory, output_file_name)
    # TODO: the -3 doesn't work with .parquet, leads to a weird file name, but duck doesn't care.
    duckdb_file_path = output_file_path[:-3] + 'duckdb'
    return output_file_path, duckdb_file_path


def remove_existing_files(output_file_path, duckdb_file_path, overwrite):
    if overwrite:
        if os.path.exists(output_file_path):
            os.remove(output_file_path)
        if os.path.exists(duckdb_file_path):
            os.remove(duckdb_file_path)


def process_with_duckdb(
    input_file_path,
    duckdb_file_path,
    split_multipolygons,
    verbose,
    format,
    output_file_path,
):
    # new duckdb at input file path but with .duckdb
    conn = duckdb.connect(duckdb_file_path)
    c = conn.cursor()
    c.execute(f"install spatial;")
    c.execute(f"load spatial;")
    c.execute(
        f"create table buildings as (select * EXCLUDE (latitude, longitude) from '{input_file_path}');"
    )

    if verbose:
        c.execute("SELECT COUNT(*) FROM buildings")
        print(f"Original rows: {c.fetchone()[0]}")

    if split_multipolygons:
        # Fetch the multipolygons
        c.execute("SELECT * FROM buildings WHERE geometry LIKE 'MULTIPOLYGON%'")
        results = c.fetchall()
        columns = [desc[0] for desc in c.description]

        multipolygon_count = 0

        # Process each multipolygon
        for row in results:
            multipolygon_count += 1
            row_dict = dict(zip(columns, row))
            multipolygon = wkt.loads(row_dict['geometry'])

            if verbose:
                # Print the original MultiPolygon
                feature = {
                    "type": "Feature",
                    "properties": {
                        k: v for k, v in row_dict.items() if k != 'geometry'
                    },
                    "geometry": multipolygon.__geo_interface__,
                }
                print("Original MultiPolygon:")
                print(json.dumps(feature))

            for polygon in multipolygon.geoms:
                # Convert the polygon to a GeoSeries in order to project it
                polygon_projected = gpd.GeoSeries([polygon], crs="EPSG:4326").to_crs(
                    'EPSG:6933'
                )

                # Compute the new area (geopandas calculates area in square meters for projected CRS)
                new_area = polygon_projected.area.values[0]

                # Compute the centroid and encode it into a Plus Code
                centroid = polygon.centroid
                new_plus_code = olc.encode(centroid.y, centroid.x, codeLength=12)

                # Create new properties for the polygon
                properties = {k: v for k, v in row_dict.items() if k != 'geometry'}
                properties['area_in_meters'] = new_area
                properties['full_plus_code'] = new_plus_code

                if verbose:
                    # Print the new Polygon
                    feature = {
                        "type": "Feature",
                        "properties": properties,
                        "geometry": polygon.__geo_interface__,
                    }
                    print("Component Polygon:")
                    print(json.dumps(feature))

                # Insert new polygon into buildings table
                columns_str = ', '.join(
                    [f'"{k}"' for k in properties.keys()] + ['geometry']
                )
                values_str = ', '.join(
                    [
                        f"'{v}'" if isinstance(v, str) else str(v)
                        for v in properties.values()
                    ]
                    + [f"'{polygon.wkt}'"]
                )
                c.execute(
                    f"INSERT INTO buildings ({columns_str}) VALUES ({values_str})"
                )

        if verbose:
            print(f"Processed {multipolygon_count} multipolygons.")

        # Delete the original multipolygons
        c.execute("DELETE FROM buildings WHERE geometry LIKE 'MULTIPOLYGON%'")

    if verbose:
        c.execute("SELECT COUNT(*) FROM buildings")
        print(f"Output rows: {c.fetchone()[0]}")

        c.execute("SELECT COUNT(*) FROM buildings WHERE geometry LIKE 'MULTIPOLYGON%'")
        print(f"Output multipolygons: {c.fetchone()[0]}")

        c.execute("SELECT COUNT(*) FROM buildings WHERE geometry LIKE 'POLYGON%'")
        print(f"Output polygons: {c.fetchone()[0]}")

    if format == 'fgb':
        c.execute(
            f"COPY (SELECT * EXCLUDE geometry, ST_AsWKB(ST_GeomFromText(geometry)) AS geometry from buildings) \
                TO '{output_file_path}' WITH  (FORMAT GDAL, DRIVER 'FlatGeobuf');"
        )
    elif format == 'parquet':
        c.execute(
            f"COPY (SELECT * EXCLUDE geometry, ST_AsWKB(ST_GeomFromText(geometry)) AS geometry from buildings) \
                TO '{output_file_path}' WITH  (FORMAT PARQUET, COMPRESSION '{PARQUET_COMPRESSION}');"
        )
        if RUN_GPQ_CONVERSION:
            print(
                f"Running gpq convert on {output_file_path}. This takes extra time but ensures the output is valid GeoParquet."
            )
            base_name, ext = os.path.splitext(output_file_path)
            temp_output_file_path = base_name + '_temp' + ext

            # convert from parquet file with a geometry column named wkb to GeoParquet
            command = ['gpq', 'convert', output_file_path, temp_output_file_path]
            gpq_start_time = time.time()
            subprocess.run(command, check=True)
            os.rename(temp_output_file_path, output_file_path)
            gpq_end_time = time.time()
            gpq_elapsed_time = gpq_end_time - gpq_start_time
            print(f"Time taken to run gpq: {gpq_elapsed_time:.2f} seconds")
        else:
            print(
                f"Skipping gpq convert on {output_file_path}. This means the output will be WKB, but it will need to be converted to GeoParquet."
            )
    elif format == 'gpkg':
        if SKIP_DUCK_GPKG:
            print(
                f"Skipping duckdb-gpkg conversion on {output_file_path}, since SKIP_DUCK_GPKG is set to True. There is likely a bug, since it takes way longer and skews the graphs"
            )
        else:
            c.execute(
                f"COPY (SELECT * EXCLUDE geometry, ST_AsWKB(ST_GeomFromText(geometry)) AS geometry from buildings) \
                    TO '{output_file_path}' WITH  (FORMAT GDAL, DRIVER 'GPKG');"
            )
    elif format == 'shp':
        c.execute(
            f"COPY (SELECT * EXCLUDE geometry, ST_AsWKB(ST_GeomFromText(geometry)) AS geometry from buildings) \
                TO '{output_file_path}' WITH  (FORMAT GDAL, DRIVER 'ESRI Shapefile');"
        )

    conn.close()


def process_with_pandas(
    input_file_path, split_multipolygons, verbose, format, output_file_path
):
    df = pd.read_csv(input_file_path)
    df['geometry'] = df['geometry'].apply(wkt.loads)

    # Drop the 'latitude' and 'longitude' columns
    df = df.drop(['latitude', 'longitude'], axis=1)

    # Convert the DataFrame to a GeoDataFrame
    gdf = gpd.GeoDataFrame(df, geometry='geometry')
    gdf.set_crs("EPSG:4326", inplace=True)

    # Create an empty GeoDataFrame for the output
    output_gdf = gpd.GeoDataFrame(columns=list(gdf.columns), crs=gdf.crs)

    if split_multipolygons:
        multipolygons = gdf[gdf.geometry.type == 'MultiPolygon']
        multipolygon_count = 0
        for i, row in multipolygons.iterrows():
            multipolygon_count += 1
            # Print the original MultiPolygon
            feature = {
                "type": "Feature",
                "properties": row.drop('geometry').to_dict(),
                "geometry": row.geometry.__geo_interface__,
            }
            if verbose:
                print("Original MultiPolygon:")
                print(json.dumps(feature))

            # Print each component Polygon
            for polygon in row.geometry.geoms:
                # Convert the polygon to a GeoSeries in order to project it
                polygon_projected = gpd.GeoSeries([polygon], crs=gdf.crs).to_crs(
                    'EPSG:6933'
                )

                # Compute the new area (geopandas calculates area in square meters for projected CRS)
                new_area = polygon_projected.area.values[0]

                # Compute the centroid and encode it into a Plus Code
                centroid = polygon.centroid
                new_plus_code = olc.encode(centroid.y, centroid.x, codeLength=12)

                # Create new properties for the polygon
                properties = row.drop('geometry').to_dict()
                properties['area_in_meters'] = new_area
                properties['full_plus_code'] = new_plus_code

                # Append to the output GeoDataFrame
                output_gdf = pd.concat(
                    [
                        output_gdf,
                        gpd.GeoDataFrame([properties], geometry=[polygon], crs=gdf.crs),
                    ],
                    ignore_index=True,
                )

                # Print the new Polygon
                feature = {
                    "type": "Feature",
                    "properties": properties,
                    "geometry": polygon.__geo_interface__,
                }
                if verbose:
                    print("Component Polygon:")
                    print(json.dumps(feature))

        print(f"Processed {multipolygon_count} multipolygons.")
        # Add the original Polygons to the output
        polygons = gdf[gdf.geometry.type == 'Polygon']
        output_gdf = pd.concat([output_gdf, polygons], ignore_index=True)
    else:
        output_gdf = gdf

    if verbose:
        # Print the number of original rows in the datafram, and the number of rows in the output
        print(f"Original rows: {len(gdf)}")
        print(f"Output rows: {len(output_gdf)}")
        # Print number of multipolygons and polygons for the output_gdf
        print(
            f"Output multipolygons: {len(output_gdf[output_gdf.geometry.type == 'MultiPolygon'])}"
        )
        print(
            f"Output polygons: {len(output_gdf[output_gdf.geometry.type == 'Polygon'])}"
        )
    # Write the output GeoDataFrame to a file
    if format == 'fgb':
        output_gdf.to_file(output_file_path, driver="FlatGeobuf")
    elif format == 'parquet':
        output_gdf.to_parquet(output_file_path, compression=PARQUET_COMPRESSION)
    elif format == 'gpkg':
        output_gdf.to_file(output_file_path, driver='GPKG')
    elif format == 'shp':
        output_gdf.to_file(output_file_path, driver='ESRI Shapefile')


def process_with_ogr2ogr(
    input_file_path, split_multipolygons, verbose, format, output_file_path
):
    # Define the SQL query to select specific columns
    table_name = os.path.splitext(os.path.basename(input_file_path))[0]

    if format == 'fgb':
        format_string = "FlatGeobuf"
    elif format == 'parquet':
        format_string = "Parquet"
    elif format == 'gpkg':
        format_string = "GPKG"
    elif format == 'shp':
        format_string = "ESRI Shapefile"

    fields_to_keep = ['confidence', 'area_in_meters', 'full_plus_code']

    # Define the ogr2ogr command
    cmd = [
        'ogr2ogr',
        '-f',
        format_string,
        '-select',
        ','.join(fields_to_keep),
        output_file_path,
        input_file_path,
        '-oo',
        'GEOM_POSSIBLE_NAMES=geometry',
        '-a_srs',
        'EPSG:4326',
    ]

    # If split_multipolygons is True, print a message and return.
    # But skip this if the output format is Shapefile, because shapefiles don't have a difference between polygons and multipolygons.
    if split_multipolygons and format != 'shp':
        print("OGR processing doesn't yet support multi polygons")
        return

    # print the ogr2ogr command that will be run
    if verbose:
        print("ogr2ogr command:")
        print(' '.join(cmd))

    # Run the command
    subprocess.run(cmd, check=True)

    if verbose:
        print(f"Converted {input_file_path} to {output_file_path} using ogr2ogr.")


def process_csv_file(
    input_file_path,
    output_directory,
    format,
    overwrite,
    process,
    split_multipolygons,
    verbose,
):
    output_file_path, duckdb_file_path = define_output_paths(
        input_file_path, output_directory, format
    )
    remove_existing_files(output_file_path, duckdb_file_path, overwrite)

    if os.path.exists(output_file_path):
        print(f'Skipping {input_file_path} as {output_file_path} already exists.')
        return
    else:
        print(
            f'Started converting {input_file_path} with {process} to {format} at {datetime.now().strftime("%Y-%m-%d %H:%M:%S")}...'
        )

    start_time = time.time()

    if process == 'duckdb':
        process_with_duckdb(
            input_file_path,
            duckdb_file_path,
            split_multipolygons,
            verbose,
            format,
            output_file_path,
        )
    elif process == 'pandas':
        process_with_pandas(
            input_file_path, split_multipolygons, verbose, format, output_file_path
        )
    elif process == 'ogr':
        process_with_ogr2ogr(
            input_file_path, split_multipolygons, verbose, format, output_file_path
        )

    execution_time = time.time() - start_time
    print(
        f'Finished processing {output_file_path} at {datetime.now().strftime("%Y-%m-%d %H:%M:%S")}. Execution time: {str(timedelta(seconds=execution_time))}'
    )


def process_geometries(
    input_path,
    output_directory,
    format,
    overwrite,
    process,
    split_multipolygons,
    verbose,
):
    # Check if the provided path is a directory or a file
    if os.path.isdir(input_path):
        # List all csv files in the directory
        csv_files = glob.glob(os.path.join(input_path, '*.csv'))

        # Sort files by size in ascending order
        csv_files.sort(key=lambda x: os.path.getsize(x))

        # Process each csv file
        for input_file_path in csv_files:
            process_csv_file(
                input_file_path,
                output_directory,
                format,
                overwrite,
                process,
                split_multipolygons,
                verbose,
            )
    elif os.path.isfile(input_path) and input_path.endswith('.csv'):
        # Process the single csv file
        process_csv_file(
            input_path,
            output_directory,
            format,
            overwrite,
            process,
            split_multipolygons,
            verbose,
        )
    else:
        raise ValueError(f"Invalid input path: {input_path}")


def process_benchmark(
    input_path, output_directory, processes, formats, split_multipolygons, verbose
):
    results = []
    for process in processes:
        for format in formats:
            start_time = time.time()
            process_geometries(
                input_path,
                output_directory,
                format,
                True,
                process,
                split_multipolygons,
                verbose,
            )
            execution_time = time.time() - start_time
            if process == 'duckdb' and format == 'gpkg' and SKIP_DUCK_GPKG:
                execution_time = 0
            results.append(
                {
                    'process': process,
                    'format': format,
                    #'execution_time': str(timedelta(seconds=execution_time)),
                    'execution_time': execution_time,
                }
            )
    return results

if __name__ == "__main__":
    cli()



# This script is used to take an Overture Parquet file and add columns
# useful for partitioning - it can put in both a quadkey and the country
# ISO code. And then it will write out parquet and use gpq to convert the
# parquet to geoparquet.


import os
import duckdb
import time
import tempfile
import subprocess
import glob
from duckdb.typing import *
import mercantile
import shutil

def lat_lon_to_quadkey(lat: DOUBLE, lon: DOUBLE, level: INTEGER) -> VARCHAR:
    # Convert latitude and longitude to tile using mercantile
    tile = mercantile.tile(lon, lat, level)
    
    # Convert the tile to a quadkey
    quadKey = mercantile.quadkey(tile)
    return quadKey

def midpoint(minval: DOUBLE, maxval: DOUBLE) -> DOUBLE:
    return (minval + maxval) / 2.0

def add_quadkey(con):

    # Register Python UDFs
    con.create_function('lat_lon_to_quadkey', lat_lon_to_quadkey, [DOUBLE, DOUBLE, INTEGER], VARCHAR)
    con.create_function('midpoint', midpoint, [DOUBLE, DOUBLE], DOUBLE)

    # Add a quadkey column to the table if it doesn't exist
    con.execute("ALTER TABLE buildings ADD COLUMN IF NOT EXISTS quadkey VARCHAR")

    # Update the quadkey column
    con.execute("""
    UPDATE buildings 
    SET quadkey = lat_lon_to_quadkey(
        midpoint(bbox.miny, bbox.maxy), 
        midpoint(bbox.minx, bbox.maxx), 
        12
    );
    """)

def add_country_iso(con, country_parquet_path):
    # Load country parquet file into duckdb
    con.execute(f"CREATE TABLE countries AS SELECT * FROM read_parquet('{country_parquet_path}')")

    # Add a country_iso column to the buildings table
    con.execute("ALTER TABLE buildings ADD COLUMN IF NOT EXISTS country_iso VARCHAR")
    
    # Update the country_iso column in the buildings table
    con.execute("""
    UPDATE buildings 
    SET country_iso = countries.isocountrycodealpha2 
    FROM countries 
    WHERE ST_Intersects(ST_GeomFromWKB(countries.geometry), ST_GeomFromWKB(buildings.geometry))
    """)

def process_parquet_file(input_parquet_path, output_folder, country_parquet_path, overwrite=False, add_quadkey_option=False, add_country_iso_option=False, verbose=False):
    # Ensure output_folder exists
    os.makedirs(output_folder, exist_ok=True)
    
    # Get unique identifier from file name
    unique_id = os.path.basename(input_parquet_path).split('_')[-1]
    
    # Define output paths
    output_db_path = os.path.join(output_folder, f'{unique_id}.duckdb')
    output_parquet_path = os.path.join(output_folder, f'{unique_id}.parquet')
    
    # Check if output files exist
    if (os.path.exists(output_db_path) or os.path.exists(output_parquet_path)) and not overwrite:
        print(f'Files with ID {unique_id} already exist. Skipping...')
        return
    
    # Overwrite mode: remove existing files
    if overwrite:
        for file_path in [output_db_path, output_parquet_path]:
            if os.path.exists(file_path):
                os.remove(file_path)
    timestamp = time.time()
    print(f"Starting processing for file {input_parquet_path} at {time.ctime(timestamp)}")
    
    # Connect to DuckDB
    con = duckdb.connect(output_db_path)
    
    con.execute('LOAD spatial;')

    # Load parquet file into duckdb
    con.execute(f"CREATE TABLE buildings AS SELECT * FROM read_parquet('{input_parquet_path}')")
    
    if add_quadkey_option:
        add_quadkey(con)

    if add_country_iso_option:
        add_country_iso(con, country_parquet_path)

    # Write out to Parquet
    con.execute(f"COPY (SELECT * FROM buildings ORDER BY quadkey) TO '{output_parquet_path}' WITH (FORMAT Parquet)")
    
    #TODO: turn this into an option to convert to geoparquet or not
    if (True):
        print(f"Converting to geoparquet: {output_parquet_path}")
        # Create a temporary file
        temp_file = tempfile.NamedTemporaryFile(suffix=".parquet", delete=False)
        temp_file.close()  # Close the file so gpq can open it

        # Convert the Parquet file to a GeoParquet file using gpq
        gpq_cmd = ['gpq', 'convert', f'{output_parquet_path}', temp_file.name]
        subprocess.run(gpq_cmd, check=True)

        # Rename the temp file to the final filename
        shutil.move(temp_file.name, f'{output_parquet_path}')
        #os.rename(temp_file.name, f'{output_parquet_path}')

    print(f"Processing complete for file {input_parquet_path}")

def process_parquet_files(input_path, output_folder, country_parquet_path, overwrite=False, add_quadkey_option=False, add_country_iso_option=False, verbose=False):
    # If input_path is a directory, process all Parquet files in it
    if os.path.isdir(input_path):
        for file in glob.glob(os.path.join(input_path, "*")):
            process_parquet_file(file, output_folder, country_parquet_path, overwrite, add_quadkey_option, add_country_iso_option, verbose)
    else:
        process_parquet_file(input_path, output_folder, country_parquet_path, overwrite, add_quadkey_option, add_country_iso_option, verbose)

# Call the function - uncomment if you want to call this directly from python and put values in here.
#input_path = '/Volumes/fastdata/overture/s3-data/buildings/'
#output_folder = '/Volumes/fastdata/overture/refined-parquet/'
#country_parquet_path = '/Volumes/fastdata/overture/countries.parquet'
#process_parquet_files(input_path, output_folder, country_parquet_path, overwrite=False, add_quadkey_option=True, add_country_iso_option=True)


import json
import click
from math import tan, cos, log, pi
from shapely.geometry import shape
from typing import Tuple
import mercantile 
import duckdb
import time
import datetime
import os
import pandas as pd
import geopandas as gpd
import subprocess
from shapely import wkb
import shutil


def geojson_to_quadkey(data: dict) -> str:
    if 'bbox' in data:
        min_lon, min_lat, max_lon, max_lat = data['bbox']
    else:
        coords = data['geometry']['coordinates'][0]
        min_lon = min_lat = float('inf')
        max_lon = max_lat = float('-inf')
        
        for lon, lat in coords:
            min_lon = min(min_lon, lon)
            min_lat = min(min_lat, lat)
            max_lon = max(max_lon, lon)
            max_lat = max(max_lat, lat)

    for zoom in range(12, -1, -1):
        tiles = list(mercantile.tiles(min_lon, min_lat, max_lon, max_lat, zooms=zoom))
        if len(tiles) == 1:
            return mercantile.quadkey(tiles[0])

    return ''

def geojson_to_wkt(data: dict) -> str:
    geometry = shape(data['geometry'])
    return geometry.wkt

def quadkey_to_geojson(quadkey: str) -> dict:
    # Convert the quadkey to tile coordinates
    tile = mercantile.quadkey_to_tile(quadkey)
    
    # Get the bounding box of the tile
    bbox = mercantile.bounds(tile)
    
    # Construct a GeoJSON Polygon representation of the bounding box
    geojson = {
        "type": "Feature",
        "geometry": {
            "type": "Polygon",
            "coordinates": [[
                [bbox.west, bbox.south],
                [bbox.east, bbox.south],
                [bbox.east, bbox.north],
                [bbox.west, bbox.north],
                [bbox.west, bbox.south]
            ]]
        }
    }
    
    return geojson

@click.group()
def cli():
    pass

@cli.command()
@click.argument('geojson_input', type=click.File('r'), required=False)
def quadkey(geojson_input):
    """Convert GeoJSON to quadkey."""
    if geojson_input:
        geojson_data = json.load(geojson_input)
    else:
        geojson_data = json.load(click.get_text_stream('stdin'))
    
    result = geojson_to_quadkey(geojson_data)
    click.echo(result)

@cli.command()
@click.argument('geojson_input', type=click.File('r'), required=False)
def WKT(geojson_input):
    """Convert GeoJSON to Well Known Text."""
    if geojson_input:
        geojson_data = json.load(geojson_input)
    else:
        geojson_data = json.load(click.get_text_stream('stdin'))
    
    result = geojson_to_wkt(geojson_data)
    click.echo(result)


@click.command()
@click.argument('geojson_input', type=click.File('r'), required=False)
@click.option('--only-quadkey', is_flag=True, help='Include only the quadkey in the WHERE clause.')
@click.option('--local', is_flag=True, help='Use local path for parquet files instead of the S3 URL.')
def sql(geojson_input, only_quadkey, local):
    """Generate an SQL query based on the input GeoJSON."""
    
    # Read the GeoJSON
    if geojson_input:
        geojson_data = json.load(geojson_input)
    else:
        geojson_data = json.load(click.get_text_stream('stdin'))

    quadkey = geojson_to_quadkey(geojson_data)
    wkt = geojson_to_wkt(geojson_data)

    # Adjust the path in read_parquet based on the --local flag
    path = '*.parquet' if local else 's3://us-west-2.opendata.source.coop/cholmes/overture/geoparquet-country-quad-2/*.parquet'
    base_sql = f"select * from read_parquet('{path}')"
    
    # Construct the WHERE clause based on the options
    where_clause = f"WHERE quadkey LIKE '{quadkey}%'"
    if not only_quadkey:
        where_clause += f" AND\nST_Within(ST_GeomFromWKB(geometry), ST_GeomFromText('{wkt}'))"

    sql_query = f"{base_sql},\n{where_clause}"
    full_sql_query = f"COPY ('{sql_query}' TO 'buildings.fgb' WITH (FORMAT GDAL, DRIVER 'FlatGeobuf')"
    click.echo(full_sql_query) 

@cli.command()
@click.argument('quadkey_input', type=str)
def quad2json(quadkey_input):
    """Convert quadkey to GeoJSON."""
    result = quadkey_to_geojson(quadkey_input)
    click.echo(json.dumps(result, indent=2))


def download(geojson_input, format, generate_sql, dst, silent, overwrite, verbose, data_path, hive_partitioning, country_iso):

    def print_timestamped_message(message):
        if not silent:
            current_time = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
            click.echo(f"[{current_time}] {message}")

    start_time = time.time()
    if verbose:
        print_timestamped_message("Reading GeoJSON input...")

    # Read the GeoJSON
    if geojson_input:
        geojson_data = json.load(geojson_input)
    else:
        geojson_data = json.load(click.get_text_stream('stdin'))

    if os.path.exists(dst) and not generate_sql:
        if overwrite:
            if verbose:
                print_timestamped_message(f"Deleting existing file at {dst}.")
            os.remove(dst)
        else:
            # Print message that the file already exists and cleanly exit the program
            print_timestamped_message(f"File at {dst} already exists. Use --overwrite to overwrite it.")
            return

    if verbose:
        print_timestamped_message("Converting GeoJSON to quadkey and WKT...")
    quadkey = geojson_to_quadkey(geojson_data)
    wkt = geojson_to_wkt(geojson_data)

    country_info = ""
    if country_iso is not None:
        country_info = f"in country {country_iso}"
    print_timestamped_message(f"Querying and downloading data for quadkey {quadkey} {country_info}...")
    if verbose:
        print_timestamped_message(f"WKT: {wkt}")
    if country_info != "":
        print_timestamped_message(f"Expect query times of at least 5-10 seconds")
    else:
        print_timestamped_message(f"Expect query times of at least 30 seconds - this can be lessened by using the --country-iso option")
   
    hive_value = 1 if hive_partitioning else 0
    select_values = "* EXCLUDE geometry"
    # if data path is overture and the output is not parquet, then name the values to get
    # so we don't get the crazy structs that gis formats barf on
    if data_path == "s3://us-west-2.opendata.source.coop/cholmes/overture/geoparquet-country-quad-hive/*/*.parquet" and format != "parquet":
        select_values = "id, level, height, numfloors, class, country_iso, quadkey"
    base_sql = f"select {select_values}, ST_AsWKB(ST_GeomFromWKB(geometry)) AS geometry from read_parquet('{data_path}', hive_partitioning={hive_value})"
    where_clause = "WHERE "
    if country_iso:
        where_clause += f"country_iso = '{country_iso}' AND "
    where_clause += f"quadkey LIKE '{quadkey}%'"
    where_clause += f" AND\nST_Within(ST_GeomFromWKB(geometry), ST_GeomFromText('{wkt}'))"

    output_extension = {
        'shapefile': '.shp',
        'geojson': 'json',
        'geopackage': '.gpkg',
        'flatgeobuf': '.fgb',
        'parquet': '.parquet'
    }

    if not format:
        for fmt, ext in output_extension.items():
            if dst.endswith(ext):
                format = fmt
                break
        else:  # The for-else structure means the else block runs if the loop completes normally, without a break.
            raise ValueError("Unknown file format. Please specify using --format option.")

    if not dst.endswith(output_extension[format]):
        dst += output_extension[format]

    create_clause = f"CREATE TABLE buildings AS ({base_sql},\n{where_clause});"
    if generate_sql or verbose:
        print_timestamped_message(create_clause)
    if not generate_sql:
        conn = duckdb.connect(database=':memory:')
        conn.execute("load spatial;")
        conn.execute(create_clause)

        count = conn.execute("SELECT COUNT(*) FROM buildings;").fetchone()[0]

        print_timestamped_message(f"Downloaded {count} features into DuckDB.")
    if not generate_sql:
        print_timestamped_message(f"Writing to {dst}...")

    if format == 'parquet':
        copy_statement = f"COPY buildings TO '{dst}' WITH (FORMAT Parquet);"
        if generate_sql or verbose:
            print_timestamped_message(copy_statement)
        if not generate_sql:
            conn.execute(f"COPY buildings TO '{dst}' WITH (FORMAT Parquet);")
            try:
                df = pd.read_parquet(dst)

                # Convert WKB geometry to geopandas geometry
                df['geometry'] = df['geometry'].apply(wkb.loads, hex=True)
                gdf = gpd.GeoDataFrame(df, geometry='geometry', crs="EPSG:4326")
                # Change output file the input_filename with .parquet replaced with _geo.parquet
                output_filename = dst.replace(".parquet", "_geo.parquet")
            
                gdf.to_parquet(output_filename)
                # delete the original file
                os.remove(dst)
                # Rename (move) the output file to the input filename
                shutil.move(output_filename, dst)
                if verbose:
                    print_timestamped_message(f"Finished processing {dst} at {time.ctime()}")
            except Exception as e:
                print(f"Error processing {dst} to geoparquet: {e}")
    else:
        gdal_format = {
            'shapefile': 'ESRI Shapefile',
            'geojson': 'GeoJSON',
            'geopackage': 'GPKG',
            'flatgeobuf': 'FlatGeobuf'
        }
        conn.execute(f"COPY buildings TO '{dst}' WITH (FORMAT GDAL, DRIVER '{gdal_format[format]}');")
    end_time = time.time()

    if verbose:
        elapsed_time = end_time - start_time
        print_timestamped_message(f"Operation took {elapsed_time:.2f} seconds.")      


# Registering the commands with the main group
cli.add_command(quadkey)
cli.add_command(WKT)
cli.add_command(sql)
cli.add_command(quad2json)
#cli.add_command(download)

if __name__ == '__main__':
    cli()




# WARNING - Work in progress
# This isn't working yet, but it's close. The main issue is that the catalog
# and collections aren't getting formed right - I want them in the hive partitions, but
# pystac keeps trying to move them in the recommended STAC structure. Committing in case 
# its useful.
# Next approach may just be to form the items individually, as that part seems to be fine,
# and then place them in the catalog and collection manually (maybe pystac can help, but 
# may be easier to just use python to adjust the links)


import os
import pystac
from pystac import Catalog, Collection, Item, Asset, CatalogType
import geopandas as gpd
from datetime import datetime
import click
from shapely.geometry import box
from dateutil.parser import parse

def read_geoparquet_bounds(filepath):
    """
    Reads a Geoparquet file and returns its bounds and EPSG.
    """
    gdf = gpd.read_parquet(filepath)
    bounds = gdf.total_bounds.tolist()
    epsg = gdf.crs.to_epsg()  # Extract the EPSG code
    return bounds, epsg

def create_stac_item_for_geoparquet(filepath, collection, item_datetime):
    filename = os.path.basename(filepath)
    file_id, _ = os.path.splitext(filename)
    title = filename

    # Get the bounds and CRS
    bbox, epsg = read_geoparquet_bounds(filepath)
    
    # Use the bounds as the geometry too
    geometry = box(*bbox).__geo_interface__

    item = Item(id=file_id,
                geometry=geometry, 
                bbox=bbox,
                datetime=item_datetime,
                properties={'title': title, 'proj:epsg': epsg},
                collection=collection.id)

    pystac.extensions.projection.ProjectionExtension.add_to(item)
    item.add_asset(key="data", asset=Asset(href=filepath, media_type="application/parquet"))

    return item

@click.command()
@click.argument('directory', type=click.Path(exists=True))
@click.option('--collection-path', default='collection.json', help='Path to the collection.json file relative to the directory.')
@click.option('--item-datetime', default='2023-05-30T00:00:00Z', help='Datetime for the STAC items.')
@click.option('--catalog-type', type=click.Choice(['SELF_CONTAINED', 'ABSOLUTE_PUBLISHED'], case_sensitive=False), default='SELF_CONTAINED', help='Type of the catalog.')
@click.option('--root-path', default=None, help='Root path for the catalog. Relevant for ABSOLUTE_PUBLISHED catalog type.')
# ... [other necessary imports and functions]

def main(directory, collection_path, item_datetime, catalog_type, root_path):
    catalog_id = 'my-catalog'
    catalog_description = 'A catalog of geoparquet files.'
    item_datetime = parse(item_datetime)
    collection = Collection.from_file(collection_path)
    
    # Create the catalog first
    catalog = Catalog(id=catalog_id, description=catalog_description, catalog_type=CatalogType[catalog_type])
    
    items = []
    for root, _, files in os.walk(directory):
        for filename in files:
            if filename.endswith(".parquet"):
                filepath = os.path.join(root, filename)
                item = create_stac_item_for_geoparquet(filepath, collection, item_datetime)
                
                # Save the item alongside the parquet file
                item_path = os.path.join(root, f"{item.id}.json")
                item.set_self_href(item_path)
                item.save_object()
                items.append(item)
    
     # Create and save the catalog
    catalog_path = os.path.join(directory, 'catalog.json')
    catalog.set_self_href(catalog_path)
    catalog.save_object()

    # Reload the catalog from file
    catalog = Catalog.from_file(catalog_path)

    # Add items to the catalog
    for item in items:
        catalog.add_item(item)
        item.add_link(pystac.Link("parent", os.path.relpath(catalog.get_self_href(), os.path.dirname(item.get_self_href()))))

    # Save the updated catalog
    catalog.save_object()

    # Load the collection and set its links
    collection_path_new = os.path.join(directory, "collection.json")
    collection.set_self_href(collection_path_new)
    collection.add_child(catalog)
    collection.save_object()

if __name__ == "__main__":
    main()



import sys
import os
import click
import pandas as pd
import matplotlib.pyplot as plt
from open_buildings.google.process import process_benchmark, process_geometries
from open_buildings.download_buildings import download as download_buildings
from open_buildings.overture.add_columns import process_parquet_files
from open_buildings.overture.partition import process_db
from datetime import datetime, timedelta
from tabulate import tabulate
import boto3  # Required for S3 operations

@click.group()
def main():
    """CLI for Open Buildings operations."""
    pass

@click.group()
def google():
    """Commands related to Google operations."""
    pass

@click.group()
def overture():
    """Commands related to Overture operations."""
    pass

main.add_command(google)
main.add_command(overture)

def handle_comma_separated(ctx, param, value):
    return value.split(',')

@main.command(name="get_buildings")
@click.argument('geojson_input', type=click.File('r'), required=False)
@click.argument('dst', type=str, default="buildings.json")
@click.option('--source', default="overture", type=click.Choice(['google', 'overture']), help='Dataset to query, defaults to Overture')
@click.option('--country_iso', type=str, default=None, help='A 2 character country ISO code to filter the data by.')
@click.option('-s', '--silent', is_flag=True, default=False, help='Suppress all print outputs.')
@click.option('--overwrite', default=False, is_flag=True, help='Overwrite the destination file if it already exists.')
@click.option('--verbose', default=False, is_flag=True, help='Print detailed logs with timestamps.')
def get_buildings(geojson_input, dst, source, country_iso, silent, overwrite, verbose):
    """Tool to extract buildings in common geospatial formats from large archives of GeoParquet data online. GeoJSON
    input can be provided as a file or piped in from stdin. If no GeoJSON input is provided, the tool will read from stdin.

    Right now the tool supports two sources of data: Google and Overture. The data comes from Cloud-Native Geospatial distributions
    on https://source.coop, that are partitioned by admin boundaries and use a quadkey for the spatial index. In time this tool will generalize
    to support any admin boundary partitioned GeoParquet data, but for now it is limited to the Google and Overture datasets.

    The default output is GeoJSON, in a file called buildings.json. Changing the suffix will change the output format - .shp for shapefile
    .gpkg for GeoPackage, .fgb for FlatGeobuf and .parquet for GeoParquet, and .json or .geojson for GeoJSON. If your query is
    all within one country it is strongly recommended to use country_iso to hint to the query engine which country to query, as this 
    will speed up the query significantly (5-10x). Expect query times of 5-10 seconds for small queries with country_iso and 30-60 seconds without country_iso.
    Large queries will take longer, as they have to download more data. 

    You can look up the country_iso for a country here: https://github.com/lukes/ISO-3166-Countries-with-Regional-Codes/blob/master/all/all.csv
    If you get the country wrong you will get zero results. Currently you can only query one country, so if your query crosses country boundaries you should
    not use country_iso. In future versions of this tool we hope to eliminate the need to hint with the country_iso.
    """
    # map source of google and overture to values for data_path and hive
    data_path = None
    hive_partitioning = False
    # case insensitive matching
    if source.lower() == "google":
        data_path = "s3://us-west-2.opendata.source.coop/google-research-open-buildings/geoparquet-by-country/*/*.parquet"
        hive_partitioning = True
    elif source.lower() == "overture":
        data_path = "s3://us-west-2.opendata.source.coop/cholmes/overture/geoparquet-country-quad-hive/*/*.parquet"
        hive_partitioning = True
    else:
        raise ValueError('Invalid source')
    
    format = None # will be set by the extension of the dst file
    generate_sql = False
    download_buildings(geojson_input, format, generate_sql, dst, silent, overwrite, verbose, data_path, hive_partitioning, country_iso)

@google.command('benchmark')
@click.argument('input_path', type=click.Path(exists=True))
@click.argument('output_directory', type=click.Path(exists=True))
@click.option(
    '--processes',
    callback=handle_comma_separated,
    default='duckdb,pandas,ogr',
    help="The processing methods to use. One or more of duckdb, pandas or ogr, in a comma-separated list. Default is duckdb,pandas,ogr.",
)
@click.option(
    '--formats',
    callback=handle_comma_separated,
    default='fgb,parquet,shp,gpkg',
    help="The output formats to benchmark. One or more of fgb, parquet, shp or gpkg, in a comma-separated list. Default is fgb,parquet,shp,gpkg.",
)
@click.option(
    '--skip-split-multis',
    is_flag=True,
    help="Whether to keep multipolygons as they are without splitting into their component polygons.",
)
@click.option('--no-gpq', is_flag=True, help="Disable GPQ conversion. Timing will be faster, but not valid GeoParquet (until DuckDB adds support)")
@click.option(
    '--verbose', is_flag=True, help="Whether to print detailed processing information."
)
@click.option(
    '--output-format',
    callback=handle_comma_separated,
    default='ascii',
    help="The format of the output. Options: ascii, csv, json, chart.",
)
def benchmark(
    input_path,
    output_directory,
    processes,
    formats,
    skip_split_multis,
    no_gpq,
    verbose,
    output_format,
):
    """Runs the convert function on each of the supplied processes and formats, printing the timing of each as a table"""
    results = process_benchmark(
        input_path, output_directory, processes, formats, not skip_split_multis, verbose
    )

    df = pd.DataFrame(results)
    df = df.pivot(index='process', columns='format', values='execution_time')

    base_name = os.path.basename(input_path)
    file_name, file_ext = os.path.splitext(base_name)

    for format in output_format:
        if format == 'csv':
            df.to_csv(f"{output_directory}/{file_name}_benchmark.csv", index=False)
        elif format == 'json':
            df.to_json(f"{output_directory}/{file_name}_benchmark.json", orient='split', indent=4)
        elif format == 'chart':
            df.plot(kind='bar', rot=0)
            plt.title(f'Benchmark for file: {base_name}')
            plt.xlabel('Process')
            plt.ylabel('Execution Time (in seconds)')
            plt.tight_layout()
            plt.savefig(f"{output_directory}/{file_name}_benchmark.png")
            plt.clf()
        elif format == 'ascii':
            df_formatted = df.copy()
            for column in df_formatted.columns:
                df_formatted[column] = df_formatted[column].apply(lambda x: (datetime.min + timedelta(seconds=x)).strftime('%M:%S.%f')[:-3])

            print(f"\nTable for file: {base_name}")
            print(tabulate(df_formatted, headers="keys", tablefmt="fancy_grid"))
        else:
            raise ValueError('Invalid output format')

@google.command('convert')
@click.argument('input_path', type=click.Path(exists=True))
@click.argument('output_directory', type=click.Path(exists=True))
@click.option(
    '--format',
    type=click.Choice(['fgb', 'parquet', 'gpkg', 'shp']),
    default='fgb',
    help="The output format. The default is FlatGeobuf (fgb)",
)
@click.option(
    '--overwrite', is_flag=True, help="Whether to overwrite any existing output files."
)
@click.option(
    '--process',
    type=click.Choice(['duckdb', 'pandas', 'ogr']),
    default='pandas',
    help="The processing method to use. The default is pandas.",
)
@click.option(
    '--skip-split-multis',
    is_flag=True,
    help="Whether to keep multipolygons as they are without splitting into their component polygons.",
)
@click.option(
    '--verbose', is_flag=True, help="Whether to print detailed processing information."
)
def convert(
    input_path, output_directory, format, overwrite, process, skip_split_multis, verbose
):
    """Converts a CSV or a directory of CSV's to an alternate format. Input CSV's are assumed to be from Google's Open Buildings"""
    process_geometries(
        input_path,
        output_directory,
        format,
        overwrite,
        process,
        not skip_split_multis,
        verbose,
    )

@overture.command('add_columns')
@click.argument('input_folder', type=click.Path(exists=True))
@click.argument('output_folder', type=click.Path())
@click.argument('country_parquet_path', type=click.Path(exists=True))
@click.option('--overwrite', is_flag=True, help="Whether to overwrite any existing output files.")
@click.option('--no-quadkey', is_flag=True, help="Whether to add a quadkey column to the output.")
@click.option('--no-country-iso', is_flag=True, help="Whether to add a country_iso column to the output.")
@click.option('--verbose', is_flag=True, help="Whether to print detailed processing information.")
def add_columns(
    input_folder, output_folder, country_parquet_path, overwrite, no_quadkey, no_country_iso, verbose
):
    """Adds columns to the input Overture parquet files, using Overture country for admin boundaries, outputting GeoParquet ordered by quadkey the output folder"""
    add_quadkey = not no_quadkey
    add_country_iso = not no_country_iso
    """Adds columns to the input parquet files, outputting to the output folder"""
    process_parquet_files(
        input_folder, output_folder, country_parquet_path, overwrite, add_quadkey, add_country_iso, verbose
    )

@overture.command('download')
@click.argument('destination_folder', type=click.Path())
@click.option(
    '--theme',
    type=click.Choice(['buildings', 'admins', 'places', 'transportation']),
    default='buildings',
    help="Theme option for the files to download from S3. Default is buildings.",
)
def overture_download(destination_folder, theme):
    """Download building files from S3 (can change theme for other overture data)."""

    os.makedirs(destination_folder, exist_ok=True)

    s3 = boto3.client('s3')
    bucket = 'overturemaps-us-west-2'
    prefix = f"release/2023-07-26-alpha.0/theme={theme}/"
    
    objects = s3.list_objects(Bucket=bucket, Prefix=prefix)
    
    for obj in objects.get('Contents', []):
        print
        file_name = os.path.basename(obj['Key'])
        local_file_path = os.path.join(destination_folder, file_name)
        timestamp = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        print(f"[{timestamp}] Downloading {file_name} to {destination_folder}")
        s3.download_file(bucket, obj['Key'], local_file_path)

        timestamp = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        print(f"[{timestamp}] Downloaded {file_name}")

@overture.command('partition')
@click.argument('duckdb-path', type=click.Path(exists=True))
@click.option('--output-folder', default=os.getcwd(), type=click.Path(), help='Folder to store the output files')
@click.option('--geo-conversion', default='gpq', type=click.Choice(['gpq', 'none', 'pandas', 'ogr'], case_sensitive=False))
@click.option('--verbose', is_flag=True, default=False, help='Print verbose output')
@click.option('--max-per-file', default=10000000, type=int, help='Maximum number of rows per file')
@click.option('--row-group-size', default=10000, type=int, help='Row group size for Parquet files')
@click.option('--hive', is_flag=True, default=False, help='Output files in Hive format (folder structure)')
@click.option('--table-name', default='buildings', type=str, help='Name of the table to process')
def partition(duckdb_path, output_folder, geo_conversion, verbose, max_per_file, row_group_size, hive, table_name):
    """Partition a DuckDB database of all overture data by country_iso"""
    process_db(duckdb_path, output_folder, geo_conversion, verbose, max_per_file, row_group_size, hive, table_name)


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python

"""Tests for `open_buildings` package."""


import unittest
from click.testing import CliRunner

from open_buildings.google.process import process_benchmark, process_geometries
from open_buildings import cli


class TestOpen_buildings(unittest.TestCase):
    """Tests for `open_buildings` package."""

    def setUp(self):
        """Set up test fixtures, if any."""

    def tearDown(self):
        """Tear down test fixtures, if any."""

    def test_000_something(self):
        """Test something."""

    # def test_command_line_interface(self):
    #     """Test the CLI."""
    #     runner = CliRunner()
    #     result = runner.invoke(cli.main)
    #     assert result.exit_code == 0
    #     assert 'open_buildings.cli.main' in result.output
    #     help_result = runner.invoke(cli.main, ['--help'])
    #     assert help_result.exit_code == 0
    #     assert '--help  Show this message and exit.' in help_result.output



# This script is a slightly more generic version of the overture add_columns.py. There's
# some chance it could be completely generic, but I was just trying to work on google buildings
# so put it under there. The main difference is that it doesn't use the midpoint of the 
# bbox struct, since that's unique to overture. It just uses the centroid of the geometry.
# That could likely work just as well if not better for overture too, so we likely can just
# get rid of that.
# The other thing that would be nice to make it truly generic is to be able to supply the 
# table name, since this should work fine with other types of data. Could also just call it
# 'features' by default, the table name doesn't really matter in these processings. Should probably check
# to be sure it works with lines and points too. So this could use clean up, also just
# removing the 'midpoint' code. 

import os
import duckdb
import time
import tempfile
import subprocess
import glob
from duckdb.typing import *
import mercantile
from shapely import wkt
import shutil

def lat_lon_to_quadkey(wkt_point: VARCHAR, level: INTEGER) -> VARCHAR:

    geom = wkt.loads(wkt_point)

    # convert geom to tile
    tile = mercantile.tile(geom.x, geom.y, level)
    
    # Convert the tile to a quadkey
    quadKey = mercantile.quadkey(tile)
    return quadKey

def midpoint(minval: DOUBLE, maxval: DOUBLE) -> DOUBLE:
    return (minval + maxval) / 2.0

def add_quadkey(con):

    # Register Python UDFs
    con.create_function('lat_lon_to_quadkey', lat_lon_to_quadkey, [VARCHAR, INTEGER], VARCHAR)
    con.create_function('midpoint', midpoint, [DOUBLE, DOUBLE], DOUBLE)

    # Add a quadkey column to the table if it doesn't exist
    con.execute("ALTER TABLE buildings ADD COLUMN IF NOT EXISTS quadkey VARCHAR")

    # Update the quadkey column
    con.execute("""
    UPDATE buildings 
    SET quadkey = lat_lon_to_quadkey(ST_Centroid(ST_GeomFromWKB(geometry)),  
        12
    );
    """)

def add_country_iso(con, country_parquet_path):
    # Load country parquet file into duckdb
    con.execute(f"CREATE TABLE countries AS SELECT * FROM read_parquet('{country_parquet_path}')")

    # Add a country_iso column to the buildings table
    con.execute("ALTER TABLE buildings ADD COLUMN IF NOT EXISTS country_iso VARCHAR")
    
    # Update the country_iso column in the buildings table
    con.execute("""
    UPDATE buildings 
    SET country_iso = countries.isocountrycodealpha2 
    FROM countries 
    WHERE ST_Intersects(ST_GeomFromWKB(countries.geometry), ST_GeomFromWKB(buildings.geometry))
    """)

def process_parquet_file(input_parquet_path, output_folder, country_parquet_path, overwrite=False, add_quadkey_option=False, add_country_iso_option=False):
    # Ensure output_folder exists
    os.makedirs(output_folder, exist_ok=True)
    
    # Get unique identifier from file name
    file_id = os.path.basename(input_parquet_path)
    
    # Define output paths
    output_db_path = os.path.join(output_folder, f'{file_id}.duckdb')
    output_parquet_path = os.path.join(output_folder, f'{file_id}')
    
    # Check if output files exist
    if (os.path.exists(output_db_path) or os.path.exists(output_parquet_path)) and not overwrite:
        print(f'Files with ID {file_id} already exist. Skipping...')
        return
    
    # Overwrite mode: remove existing files
    if overwrite:
        for file_path in [output_db_path, output_parquet_path]:
            if os.path.exists(file_path):
                os.remove(file_path)
    timestamp = time.time()
    print(f"Starting processing for file {input_parquet_path} at {time.ctime(timestamp)}")
    
    # Connect to DuckDB
    con = duckdb.connect(output_db_path)
    
    con.execute('LOAD spatial;')

    # Load parquet file into duckdb
    con.execute(f"CREATE TABLE buildings AS SELECT * FROM read_parquet('{input_parquet_path}')")
    
    if add_quadkey_option:
        add_quadkey(con)

    if add_country_iso_option:
        add_country_iso(con, country_parquet_path)

    # Write out to Parquet
    con.execute(f"COPY (SELECT * FROM buildings ORDER BY quadkey) TO '{output_parquet_path}' WITH (FORMAT Parquet)")
    
    if (False):
        print(f"Converting to geoparquet: {output_parquet_path}")
        # Create a temporary file
        temp_file = tempfile.NamedTemporaryFile(suffix=".parquet", delete=False)
        temp_file.close()  # Close the file so gpq can open it

        # Convert the Parquet file to a GeoParquet file using gpq
        gpq_cmd = ['gpq', 'convert', f'{output_parquet_path}', temp_file.name]
        subprocess.run(gpq_cmd, check=True)

        # Rename the temp file to the final filename
        shutil.move(temp_file.name, f'{output_parquet_path}')
        #os.rename(temp_file.name, f'{output_parquet_path}')

    print(f"Processing complete for file {input_parquet_path}")

    remove_duckdb = False

    # remove duckdb file
    if (remove_duckdb):
        os.remove(output_db_path)

def process_parquet_files(input_path, output_folder, country_parquet_path, overwrite=False, add_quadkey_option=False, add_country_iso_option=False):
    # If input_path is a directory, process all Parquet files in it
    if os.path.isdir(input_path):
        for file in glob.glob(os.path.join(input_path, "*.parquet")):
            process_parquet_file(file, output_folder, country_parquet_path, overwrite, add_quadkey_option, add_country_iso_option)
    else:
        process_parquet_file(input_path, output_folder, country_parquet_path, overwrite, add_quadkey_option, add_country_iso_option)

# Call the function
input_path = '/Users/cholmes/geodata/google-buildings-v3/geoparquet/'
output_folder = '/Users/cholmes/geodata/google-buildings-v3/geoparquet-columns'
country_parquet_path = '/Volumes/fastdata/overture/countries.parquet'
process_parquet_files(input_path, output_folder, country_parquet_path, overwrite=False, add_quadkey_option=True, add_country_iso_option=True)